/*
 * OgreMax Sample Viewer and Scene Loader - Ogre3D-based viewer and code for loading and displaying .scene files
 * Copyright 2013 AND Entertainment
 *
 * This code is available under the OgreMax Free License:
 *   -You may use this code for any purpose, commercial or non-commercial.
 *   -If distributing derived works (that use this source code) in binary or source code form, 
 *    you must give the following credit in your work's end-user documentation: 
 *        "Portions of this work provided by OgreMax (www.ogremax.com)"
 *
 * AND Entertainment assumes no responsibility for any harm caused by using this code.
 * 
 * The OgreMax Sample Viewer and Scene Loader were released at www.ogremax.com 
 */


//Includes---------------------------------------------------------------------
#ifndef OGREMAX_VIEWER_NO_LUA

#include "ApplicationLuaTypes.hpp"
#include "InputSystem.hpp"

#include <OgreMath.h>
#include <OgreVector2.h>
#include <OgreVector3.h>
#include <OgrePlane.h>
#include <OgreMatrix3.h>
#include <OgreMatrix4.h>
#include <OgreQuaternion.h>
#include <OgreAxisAlignedBox.h>
#include <OgreSphere.h>
#include <OgreRay.h>
#include <OgreAnimationState.h>
#include <OgreNode.h>
#include <OgreSceneNode.h>
#include <OgreCamera.h>

#include <luabind/luabind.hpp>
#include <luabind/class.hpp>
#include <luabind/function.hpp>
#include <luabind/object.hpp>
#include <luabind/operator.hpp>
#include <luabind/adopt_policy.hpp>
#include <luabind/return_reference_to_policy.hpp>
#include <luabind/out_value_policy.hpp>

using namespace OgreMax;
using namespace luabind;


//Local classes----------------------------------------------------------------
class ScriptInputSystem
{
public:
    Keyboard* GetKeyboard() const
    {
        return &InputSystem::GetInstance().Keyboard;
    }

    Mouse* GetMouse() const
    {
        return &InputSystem::GetInstance().Mouse;
    }

    Joystick* GetJoystick() const
    {
        return &InputSystem::GetInstance().GetJoystick(0);
    }

    static ScriptInputSystem& GetInstance()
    {
        static ScriptInputSystem instance;
        return instance;
    }
};


//Local functions--------------------------------------------------------------
/**
 * Template function that returns a constant value
 */
template <class T, const T& value>
T ValueFunction()
{
    return value;
}

static Ogre::Vector3 GetHSB(const Ogre::ColourValue& color)
{
    Ogre::Real h, s, b;
    color.getHSB(&h, &s, &b);
    return Ogre::Vector3(h, s, b);
}



//Implementation---------------------------------------------------------------
void ApplicationLuaTypes::Register(LuaScript& script, const Ogre::String& moduleName)
{
    class ScriptKeys
    {
    };

    class ScriptTransformSpace
    {
    };

    class ScriptProjectionType
    {
    };

    //Initialize script types
    module(script, moduleName.c_str())
	[
        class_<Keyboard>("Keyboard")
            .def("PressedKey", &Keyboard::PressedKey)
            .def("PressingKey", &Keyboard::PressingKey)
            .def("ReleasedKey", &Keyboard::ReleasedKey),

        class_<ScriptKeys>("Keys")
            .enum_("constants")
            [
                value("ESCAPE", OIS::KC_ESCAPE),
		        value("_1", OIS::KC_1),
		        value("_2", OIS::KC_2),
		        value("_3", OIS::KC_3),
		        value("_4", OIS::KC_4),
		        value("_5", OIS::KC_5),
		        value("_6", OIS::KC_6),
		        value("_7", OIS::KC_7),
		        value("_8", OIS::KC_8),
		        value("_9", OIS::KC_9),
		        value("_0", OIS::KC_0),
		        value("MINUS", OIS::KC_MINUS),
		        value("EQUALS", OIS::KC_EQUALS),
		        value("BACK", OIS::KC_BACK),
		        value("TAB", OIS::KC_TAB),
		        value("Q", OIS::KC_Q),
		        value("W", OIS::KC_W),
		        value("E", OIS::KC_E),
		        value("R", OIS::KC_R),
		        value("T", OIS::KC_T),
		        value("Y", OIS::KC_Y),
		        value("U", OIS::KC_U),
		        value("I", OIS::KC_I),
		        value("O", OIS::KC_O),
		        value("P", OIS::KC_P),
		        value("LBRACKET", OIS::KC_LBRACKET),
		        value("RBRACKET", OIS::KC_RBRACKET),
		        value("RETURN", OIS::KC_RETURN),
		        value("LCONTROL", OIS::KC_LCONTROL),
		        value("A", OIS::KC_A),
		        value("S", OIS::KC_S),
		        value("D", OIS::KC_D),
		        value("F", OIS::KC_F),
		        value("G", OIS::KC_G),
		        value("H", OIS::KC_H),
		        value("J", OIS::KC_J),
		        value("K", OIS::KC_K),
		        value("L", OIS::KC_L),
		        value("SEMICOLON", OIS::KC_SEMICOLON),
		        value("APOSTROPHE", OIS::KC_APOSTROPHE),
		        value("GRAVE", OIS::KC_GRAVE),
		        value("LSHIFT", OIS::KC_LSHIFT),
		        value("BACKSLASH", OIS::KC_BACKSLASH),
		        value("Z", OIS::KC_Z),
		        value("X", OIS::KC_X),
		        value("C", OIS::KC_C),
		        value("V", OIS::KC_V),
		        value("B", OIS::KC_B),
		        value("N", OIS::KC_N),
		        value("M", OIS::KC_M),
		        value("COMMA", OIS::KC_COMMA),
		        value("PERIOD", OIS::KC_PERIOD),
		        value("SLASH", OIS::KC_SLASH),
		        value("RSHIFT", OIS::KC_RSHIFT),
		        value("MULTIPLY", OIS::KC_MULTIPLY),
		        value("LMENU", OIS::KC_LMENU),
		        value("SPACE", OIS::KC_SPACE),
		        value("CAPITAL", OIS::KC_CAPITAL),
		        value("F1", OIS::KC_F1),
		        value("F2", OIS::KC_F2),
		        value("F3", OIS::KC_F3),
		        value("F4", OIS::KC_F4),
		        value("F5", OIS::KC_F5),
		        value("F6", OIS::KC_F6),
		        value("F7", OIS::KC_F7),
		        value("F8", OIS::KC_F8),
		        value("F9", OIS::KC_F9),
		        value("F10", OIS::KC_F10),
		        value("NUMLOCK", OIS::KC_NUMLOCK),
		        value("SCROLL", OIS::KC_SCROLL),
		        value("NUMPAD7", OIS::KC_NUMPAD7),
		        value("NUMPAD8", OIS::KC_NUMPAD8),
		        value("NUMPAD9", OIS::KC_NUMPAD9),
		        value("SUBTRACT", OIS::KC_SUBTRACT),
		        value("NUMPAD4", OIS::KC_NUMPAD4),
		        value("NUMPAD5", OIS::KC_NUMPAD5),
		        value("NUMPAD6", OIS::KC_NUMPAD6),
		        value("ADD", OIS::KC_ADD),
		        value("NUMPAD1", OIS::KC_NUMPAD1),
		        value("NUMPAD2", OIS::KC_NUMPAD2),
		        value("NUMPAD3", OIS::KC_NUMPAD3),
		        value("NUMPAD0", OIS::KC_NUMPAD0),
		        value("DECIMAL", OIS::KC_DECIMAL),
		        value("OEM_102", OIS::KC_OEM_102),
		        value("F11", OIS::KC_F11),
		        value("F12", OIS::KC_F12),
		        value("F13", OIS::KC_F13),
		        value("F14", OIS::KC_F14),
		        value("F15", OIS::KC_F15),
		        value("KANA", OIS::KC_KANA),
		        value("ABNT_C1", OIS::KC_ABNT_C1),
		        value("CONVERT", OIS::KC_CONVERT),
		        value("NOCONVERT", OIS::KC_NOCONVERT),
		        value("YEN", OIS::KC_YEN),
		        value("ABNT_C2", OIS::KC_ABNT_C2),
		        value("NUMPADEQUALS", OIS::KC_NUMPADEQUALS),
		        value("PREVTRACK", OIS::KC_PREVTRACK),
		        value("AT", OIS::KC_AT),
		        value("COLON", OIS::KC_COLON),
		        value("UNDERLINE", OIS::KC_UNDERLINE),
		        value("KANJI", OIS::KC_KANJI),
		        value("STOP", OIS::KC_STOP),
		        value("AX", OIS::KC_AX),
		        value("UNLABELED", OIS::KC_UNLABELED),
		        value("NEXTTRACK", OIS::KC_NEXTTRACK),
		        value("NUMPADENTER", OIS::KC_NUMPADENTER),
		        value("RCONTROL", OIS::KC_RCONTROL),
		        value("MUTE", OIS::KC_MUTE),
		        value("CALCULATOR", OIS::KC_CALCULATOR),
		        value("PLAYPAUSE", OIS::KC_PLAYPAUSE),
		        value("MEDIASTOP", OIS::KC_MEDIASTOP),
		        value("VOLUMEDOWN", OIS::KC_VOLUMEDOWN),
		        value("VOLUMEUP", OIS::KC_VOLUMEUP),
		        value("WEBHOME", OIS::KC_WEBHOME),
		        value("NUMPADCOMMA", OIS::KC_NUMPADCOMMA),
		        value("DIVIDE", OIS::KC_DIVIDE),
		        value("SYSRQ", OIS::KC_SYSRQ),
		        value("RMENU", OIS::KC_RMENU),
		        value("PAUSE", OIS::KC_PAUSE),
		        value("HOME", OIS::KC_HOME),
		        value("UP", OIS::KC_UP),
		        value("PGUP", OIS::KC_PGUP),
		        value("LEFT", OIS::KC_LEFT),
		        value("RIGHT", OIS::KC_RIGHT),
		        value("END", OIS::KC_END),
		        value("DOWN", OIS::KC_DOWN),
		        value("PGDOWN", OIS::KC_PGDOWN),
		        value("INSERT", OIS::KC_INSERT),
		        value("DELETE", OIS::KC_DELETE),
		        value("LWIN", OIS::KC_LWIN),
		        value("RWIN", OIS::KC_RWIN),
		        value("APPS", OIS::KC_APPS),
		        value("POWER", OIS::KC_POWER),
		        value("SLEEP", OIS::KC_SLEEP),
		        value("WAKE", OIS::KC_WAKE),
		        value("WEBSEARCH", OIS::KC_WEBSEARCH),
		        value("WEBFAVORITES", OIS::KC_WEBFAVORITES),
		        value("WEBREFRESH", OIS::KC_WEBREFRESH),
		        value("WEBSTOP", OIS::KC_WEBSTOP),
		        value("WEBFORWARD", OIS::KC_WEBFORWARD),
		        value("WEBBACK", OIS::KC_WEBBACK),
		        value("MYCOMPUTER", OIS::KC_MYCOMPUTER),
		        value("MAIL", OIS::KC_MAIL),
		        value("MEDIASELECT", OIS::KC_MEDIASELECT)
            ],

        class_<Mouse>("Mouse")
            .def("PressedButton", &Mouse::PressedButton)
            .def("PressingButton", &Mouse::PressingButton)
            .def("ReleasedButton", &Mouse::ReleasedButton),

        class_<Joystick>("Joystick")
            .def("PressedButton", &Joystick::PressedButton)
            .def("PressingButton", &Joystick::PressingButton)
            .def("ReleasedButton", &Joystick::ReleasedButton),

        class_<ScriptInputSystem>("InputSystem")
            .property("Keyboard", &ScriptInputSystem::GetKeyboard)
            .property("Mouse", &ScriptInputSystem::GetMouse)
            .property("Joystick", &ScriptInputSystem::GetJoystick),

        class_<Ogre::Radian>("Radian")
            .def(constructor<Ogre::Real>())
            .def(constructor<const Ogre::Radian&>())
            .def("GetDegrees", &Ogre::Radian::valueDegrees)
            .def("GetRadians", &Ogre::Radian::valueRadians),

        class_<Ogre::Vector2>("Vector2")
            .def(constructor<>())
            .def(constructor<float, float>())
            .def(constructor<const Ogre::Vector2&>())
            .def_readwrite("X", &Ogre::Vector2::x)
            .def_readwrite("Y", &Ogre::Vector2::y)
            .def(const_self + other<const Ogre::Vector2&>())
            .def(const_self - other<const Ogre::Vector2&>())
            .def(const_self / other<const Ogre::Vector2&>())
            .def(const_self * other<const Ogre::Vector2&>())
            .def(const_self * float())
            .def(const_self / float())
            .def(float() * const_self)
            .def(float() / const_self)
            .def(const_self + float())
            .def(const_self - float())
            .def(float() + const_self)
            .def(float() - const_self)
            .def(- const_self)
            .def("Length", &Ogre::Vector2::length)
            .def("SquaredLength", &Ogre::Vector2::squaredLength)
            .def("Dot", &Ogre::Vector2::dotProduct)
            .def("Normalize", &Ogre::Vector2::normalise)
            .def("Cross", &Ogre::Vector2::crossProduct)
            .def("MidPoint", &Ogre::Vector2::midPoint)
            .def("Perpendicular", &Ogre::Vector2::perpendicular)
            .scope
            [
                def("Zero", ValueFunction<Ogre::Vector2, Ogre::Vector2::ZERO>),
                def("One", ValueFunction<Ogre::Vector2, Ogre::Vector2::UNIT_SCALE>),
                def("UnitX", ValueFunction<Ogre::Vector2, Ogre::Vector2::UNIT_X>),
                def("UnitY", ValueFunction<Ogre::Vector2, Ogre::Vector2::UNIT_Y>),
                def("NegativeUnitX", ValueFunction<Ogre::Vector2, Ogre::Vector2::NEGATIVE_UNIT_X>),
                def("NegativeUnitY", ValueFunction<Ogre::Vector2, Ogre::Vector2::NEGATIVE_UNIT_Y>)
            ],

		class_<Ogre::Vector3>("Vector3")
            .def(constructor<>())
            .def(constructor<float, float, float>())
            .def(constructor<const Ogre::Vector3&>())
            .def_readwrite("X", &Ogre::Vector3::x)
            .def_readwrite("Y", &Ogre::Vector3::y)
            .def_readwrite("Z", &Ogre::Vector3::z)
            .def(const_self + other<const Ogre::Vector3&>())
            .def(const_self - other<const Ogre::Vector3&>())
            .def(const_self / other<const Ogre::Vector3&>())
            .def(const_self * other<const Ogre::Vector3&>())
            .def(const_self * float())
            .def(const_self / float())
            .def(float() * const_self)
            .def(float() / const_self)
            .def(const_self + float())
            .def(const_self - float())
            .def(float() + const_self)
            .def(float() - const_self)
            .def(- const_self)
            .def("Length", &Ogre::Vector3::length)
            .def("SquaredLength", &Ogre::Vector3::squaredLength)
            .def("Distance", &Ogre::Vector3::distance)
            .def("SquaredDistance", &Ogre::Vector3::squaredDistance)
            .def("Dot", &Ogre::Vector3::dotProduct)
            .def("AbsDot", &Ogre::Vector3::absDotProduct)
            .def("Normalize", &Ogre::Vector3::normalise)
            .def("Cross", &Ogre::Vector3::crossProduct)
            .def("MidPoint", &Ogre::Vector3::midPoint)
            .def("Perpendicular", &Ogre::Vector3::perpendicular)
            .scope
            [
                def("Zero", ValueFunction<Ogre::Vector3, Ogre::Vector3::ZERO>),
                def("One", ValueFunction<Ogre::Vector3, Ogre::Vector3::UNIT_SCALE>),
                def("UnitX", ValueFunction<Ogre::Vector3, Ogre::Vector3::UNIT_X>),
                def("UnitY", ValueFunction<Ogre::Vector3, Ogre::Vector3::UNIT_Y>),
                def("UnitZ", ValueFunction<Ogre::Vector3, Ogre::Vector3::UNIT_Z>),
                def("NegativeUnitX", ValueFunction<Ogre::Vector3, Ogre::Vector3::NEGATIVE_UNIT_X>),
                def("NegativeUnitY", ValueFunction<Ogre::Vector3, Ogre::Vector3::NEGATIVE_UNIT_Y>),
                def("NegativeUnitZ", ValueFunction<Ogre::Vector3, Ogre::Vector3::NEGATIVE_UNIT_Z>)
            ],

        class_<Ogre::Plane>("Plane")
            .def(constructor<>())
            .def(constructor<const Ogre::Plane&>())
            .def(constructor<const Ogre::Vector3&, Ogre::Real>())
            .def(constructor<const Ogre::Vector3&, const Ogre::Vector3&>())
            .def(constructor<const Ogre::Vector3&, const Ogre::Vector3&, const Ogre::Vector3&>())
            .def_readwrite("Normal", &Ogre::Plane::normal)
            .def_readwrite("D", &Ogre::Plane::d)
            .def("GetSide", (Ogre::Plane::Side (Ogre::Plane::*)(const Ogre::Vector3&) const)&Ogre::Plane::getSide)
            .def("GetSide", (Ogre::Plane::Side (Ogre::Plane::*)(const Ogre::Vector3&, const Ogre::Vector3&) const)&Ogre::Plane::getSide)
            .def("GetDistance", &Ogre::Plane::getDistance)
            .def("Redefine", (void (Ogre::Plane::*)(const Ogre::Vector3&, const Ogre::Vector3&))&Ogre::Plane::redefine)
            .def("Redefine", (void (Ogre::Plane::*)(const Ogre::Vector3&, const Ogre::Vector3&, const Ogre::Vector3&))&Ogre::Plane::redefine)
            .def("ProjectVector", &Ogre::Plane::projectVector)
            .enum_("constants")
            [
                value("NO_SIDE", Ogre::Plane::NO_SIDE),
                value("POSITIVE_SIDE", Ogre::Plane::POSITIVE_SIDE),
                value("NEGATIVE_SIDE", Ogre::Plane::NEGATIVE_SIDE),
                value("BOTH_SIDES", Ogre::Plane::BOTH_SIDE)
            ],

        class_<Ogre::Matrix3>("Matrix3")
            .def(constructor<>())
            .def(constructor<const Ogre::Matrix3&>())
            .def(const_self * other<const Ogre::Vector3&>())
            .def("Transposed", &Ogre::Matrix3::Transpose)
            .def("Inverse", (Ogre::Matrix3 (Ogre::Matrix3::*)(Ogre::Real) const)&Ogre::Matrix3::Inverse)
            .def("GetColumn", &Ogre::Matrix3::GetColumn)
            .scope
            [
                def("Identity", ValueFunction<Ogre::Matrix3, Ogre::Matrix3::IDENTITY>)
            ],

        class_<Ogre::Matrix4>("Matrix4")
            .def(constructor<>())
            .def(constructor<const Ogre::Matrix4&>())
            .def(constructor<const Ogre::Quaternion&>())
            .def(const_self * other<const Ogre::Vector3&>())
            .def("SetScale", &Ogre::Matrix4::setScale)
            .def("GetRotationQuaternion", &Ogre::Matrix4::extractQuaternion)
            .def("GetTranslation", (Ogre::Vector3 (Ogre::Matrix4::*)() const)&Ogre::Matrix4::getTrans)
            .def("SetTranslation", &Ogre::Matrix4::setTrans)
            .def("ExtractQuaternion", &Ogre::Matrix4::extractQuaternion)
            .scope
            [
                def("Identity", ValueFunction<Ogre::Matrix4, Ogre::Matrix4::IDENTITY>)
            ],

        class_<Ogre::Quaternion>("Quaternion")
            .def(constructor<>())
            .def(constructor<const Ogre::Quaternion&>())
            .def(constructor<Ogre::Radian, const Ogre::Vector3&>())
            .def(constructor<const Ogre::Vector3&, const Ogre::Vector3&, const Ogre::Vector3&>())
            .def_readwrite("X", &Ogre::Quaternion::x)
            .def_readwrite("Y", &Ogre::Quaternion::y)
            .def_readwrite("Z", &Ogre::Quaternion::z)
            .def_readwrite("W", &Ogre::Quaternion::w)
            .def(const_self + other<const Ogre::Quaternion&>())
            .def(const_self - other<const Ogre::Quaternion&>())
            .def(const_self * other<const Ogre::Quaternion&>())
            .def("FromRotationMatrix", &Ogre::Quaternion::FromRotationMatrix)
            .def("FromAxes", (void (Ogre::Quaternion::*)(const Ogre::Vector3&, const Ogre::Vector3&, const Ogre::Vector3&))&Ogre::Quaternion::FromAxes)
            .def("Normalize", &Ogre::Quaternion::normalise)
            .def("Inverse", &Ogre::Quaternion::Inverse)
            .def("UnitInverse", &Ogre::Quaternion::UnitInverse)
            .def("Exp", &Ogre::Quaternion::Exp)
            .def("Log", &Ogre::Quaternion::Log)
            .scope
            [
                def("Identity", ValueFunction<Ogre::Quaternion, Ogre::Quaternion::IDENTITY>),
                def("Slerp", &Ogre::Quaternion::Slerp),
                def("Nlerp", &Ogre::Quaternion::nlerp)
            ],

         class_<Ogre::ColourValue>("ColourValue")
            .def(constructor<>())
            .def(constructor<Ogre::Real, Ogre::Real, Ogre::Real, Ogre::Real>())
            .def(constructor<Ogre::Real, Ogre::Real, Ogre::Real>())
            .def_readwrite("R", &Ogre::ColourValue::r)
            .def_readwrite("G", &Ogre::ColourValue::g)
            .def_readwrite("B", &Ogre::ColourValue::b)
            .def_readwrite("A", &Ogre::ColourValue::a)
            .def(const_self + other<const Ogre::ColourValue&>())
            .def(const_self - other<const Ogre::ColourValue&>())
            .def(const_self * other<const Ogre::ColourValue&>())
            .def(const_self * Ogre::Real())
            .def(const_self / Ogre::Real())
            .def("Saturate", &Ogre::ColourValue::saturate)
            .def("GetHSB", GetHSB)
            .def("SetHSB", &Ogre::ColourValue::setHSB)
            .scope
            [
                def("Zero", ValueFunction<Ogre::ColourValue, Ogre::ColourValue::ZERO>),
                def("Black", ValueFunction<Ogre::ColourValue, Ogre::ColourValue::Black>),
                def("White", ValueFunction<Ogre::ColourValue, Ogre::ColourValue::White>),
                def("Red", ValueFunction<Ogre::ColourValue, Ogre::ColourValue::Red>),
                def("Green", ValueFunction<Ogre::ColourValue, Ogre::ColourValue::Green>),
                def("Blue", ValueFunction<Ogre::ColourValue, Ogre::ColourValue::Blue>)
            ],

        class_<std::pair<bool, Ogre::Real> >("IntersectionResult")
            .def_readwrite("Intersects", &std::pair<bool, Ogre::Real>::first)
            .def_readwrite("Distance", &std::pair<bool, Ogre::Real>::second),

        class_<Ogre::AxisAlignedBox>("AxisAlignedBox")
            .def(constructor<>())
            .def(constructor<const Ogre::AxisAlignedBox&>())
            .def(constructor<const Ogre::Vector3&, const Ogre::Vector3&>())
            .def("Set", (void (Ogre::AxisAlignedBox::*)(const Ogre::Vector3&, const Ogre::Vector3&))&Ogre::AxisAlignedBox::setExtents)
            .property("Min", (const Ogre::Vector3& (Ogre::AxisAlignedBox::*)() const)&Ogre::AxisAlignedBox::getMinimum, (void (Ogre::AxisAlignedBox::*)(const Ogre::Vector3&))&Ogre::AxisAlignedBox::setMinimum)
            .property("Max", (const Ogre::Vector3& (Ogre::AxisAlignedBox::*)() const)&Ogre::AxisAlignedBox::getMaximum, (void (Ogre::AxisAlignedBox::*)(const Ogre::Vector3&))&Ogre::AxisAlignedBox::setMaximum)
            .def("Merge", (void (Ogre::AxisAlignedBox::*)(const Ogre::AxisAlignedBox&))&Ogre::AxisAlignedBox::merge)
            .def("Merge", (void (Ogre::AxisAlignedBox::*)(const Ogre::Vector3&))&Ogre::AxisAlignedBox::merge)
            .def("Transform", &Ogre::AxisAlignedBox::transform)
            .def("IsEmpty", &Ogre::AxisAlignedBox::isNull)
            .def("SetEmpty", &Ogre::AxisAlignedBox::setNull),

        class_<Ogre::Sphere>("Sphere")
            .def(constructor<>())
            .def(constructor<const Ogre::Sphere&>())
            .def(constructor<const Ogre::Vector3&, Ogre::Real>())
            .property("Center", &Ogre::Sphere::getCenter, &Ogre::Sphere::setCenter)
            .property("Radius", &Ogre::Sphere::getRadius, &Ogre::Sphere::setRadius)
            .def("Intersects", (bool (Ogre::Sphere::*)(const Ogre::Sphere&) const)&Ogre::Sphere::intersects)
            .def("Intersects", (bool (Ogre::Sphere::*)(const Ogre::Vector3&) const)&Ogre::Sphere::intersects),

        class_<Ogre::Ray>("Ray")
            .def(constructor<>())
            .def(constructor<const Ogre::Ray&>())
            .def(constructor<const Ogre::Vector3&, const Ogre::Vector3&>())
            .property("Origin", &Ogre::Ray::getOrigin, &Ogre::Ray::setOrigin)
            .property("Direction", &Ogre::Ray::getDirection, &Ogre::Ray::setDirection)
            .def("GetPoint", &Ogre::Ray::getPoint),

        class_<Ogre::Math>("Math")
            .scope
            [
                def("Sqrt", (Ogre::Real (*)(Ogre::Real))Ogre::Math::Sqrt),
                def("InvSqrt", Ogre::Math::InvSqrt),
                def("UnitRandom", Ogre::Math::UnitRandom),
                def("RangeRandom", Ogre::Math::UnitRandom),
                def("SymmetricRandom", Ogre::Math::SymmetricRandom),
                def("DegreesToRadians", Ogre::Math::DegreesToRadians),
                def("RadiansToDegrees", Ogre::Math::RadiansToDegrees),
                def("IsPointInTriangle2D", Ogre::Math::pointInTri2D),
                def("IsPointInTriangle3D", Ogre::Math::pointInTri3D),
                def("Intersects", (std::pair<bool, Ogre::Real> (*)(const Ogre::Ray&, const Ogre::Plane&))Ogre::Math::intersects),
                def("Intersects", (std::pair<bool, Ogre::Real> (*)(const Ogre::Ray&, const Ogre::Sphere&, bool))Ogre::Math::intersects),
                def("Intersects", (std::pair<bool, Ogre::Real> (*)(const Ogre::Ray&, const Ogre::AxisAlignedBox&))Ogre::Math::intersects),
                def("Intersects", (std::pair<bool, Ogre::Real> (*)(const Ogre::Ray&, const Ogre::Vector3&, const Ogre::Vector3&, const Ogre::Vector3&, bool, bool))Ogre::Math::intersects),
                def("Intersects", (bool (*)(const Ogre::Sphere&, const Ogre::AxisAlignedBox&))Ogre::Math::intersects),
                def("Intersects", (bool (*)(const Ogre::Plane&, const Ogre::AxisAlignedBox&))Ogre::Math::intersects),
                def("Intersects", (bool (*)(const Ogre::Sphere&, const Ogre::Plane&))Ogre::Math::intersects),
                def("GetFaceNormal", Ogre::Math::calculateBasicFaceNormal),
                def("GaussianDistribution", Ogre::Math::gaussianDistribution)
            ],

        class_<Ogre::AnimationState>("AnimationState")
            .property("Name", &Ogre::AnimationState::getAnimationName)
            .property("Enabled", &Ogre::AnimationState::getEnabled, &Ogre::AnimationState::setEnabled)
            .property("Looped", &Ogre::AnimationState::getLoop, &Ogre::AnimationState::setLoop)
            .property("Ended", &Ogre::AnimationState::hasEnded)
            .property("TimePosition", &Ogre::AnimationState::getTimePosition, &Ogre::AnimationState::setTimePosition)
            .property("Length", &Ogre::AnimationState::getLength, &Ogre::AnimationState::setLength)
            .property("Weight", &Ogre::AnimationState::getWeight, &Ogre::AnimationState::setWeight)
            .def("AddTime", &Ogre::AnimationState::addTime),

        class_<ScriptTransformSpace>("TransformSpace")
            .enum_("constants")
            [
                value("LOCAL", Ogre::Node::TS_LOCAL),
                value("PARENT", Ogre::Node::TS_PARENT),
				value("WORLD", Ogre::Node::TS_WORLD)
            ],

        class_<ScriptProjectionType>("ProjectionType")
            .enum_("constants")
            [
				value("ORTHOGRAPHIC", Ogre::PT_ORTHOGRAPHIC),
                value("PERSPECTIVE", Ogre::PT_PERSPECTIVE)
            ],

        class_<Ogre::Camera>("Camera")
            .property("Name", &Ogre::Camera::getName)
            .property("Position", &Ogre::Camera::getPosition, (void (Ogre::Camera::*)(const Ogre::Vector3&))&Ogre::Camera::setPosition)
            .property("Orientation", &Ogre::Camera::getOrientation, (void (Ogre::Camera::*)(const Ogre::Quaternion&))&Ogre::Camera::setOrientation)
            .property("NearClipDistance", &Ogre::Camera::getNearClipDistance, &Ogre::Camera::setNearClipDistance)
            .property("FarClipDistance", &Ogre::Camera::getFarClipDistance, &Ogre::Camera::setFarClipDistance)
            .property("FovY", &Ogre::Camera::getFOVy, (void (Ogre::Camera::*)(const Ogre::Radian&))&Ogre::Camera::setFOVy)
            .property("ProjectionType", &Ogre::Camera::getProjectionType, &Ogre::Camera::setProjectionType),

        class_<Ogre::SceneNode>("SceneNode")
            .property("Name", &Ogre::SceneNode::getName)
            .property("Position", &Ogre::SceneNode::getPosition, (void (Ogre::SceneNode::*)(const Ogre::Vector3&))&Ogre::SceneNode::setPosition)
            .property("Scale", &Ogre::SceneNode::getScale, (void (Ogre::SceneNode::*)(const Ogre::Vector3&))&Ogre::SceneNode::setScale)
            .property("Orientation", &Ogre::SceneNode::getOrientation, (void (Ogre::SceneNode::*)(const Ogre::Quaternion&))&Ogre::SceneNode::setOrientation)
            .property("DerivedPosition", &Ogre::SceneNode::_getDerivedPosition)
            .property("DerivedScale", &Ogre::SceneNode::_getDerivedScale)
            .property("DerivedOrientation", &Ogre::SceneNode::_getDerivedOrientation)
            .def("ResetOrientation", &Ogre::SceneNode::resetOrientation)
            .def("FlipVisibility", &Ogre::SceneNode::flipVisibility)
            .def("SetVisible", &Ogre::SceneNode::setVisible)
            .def("ShowBoundingBox", &Ogre::SceneNode::showBoundingBox)
    ];

    //Set the input system object as a global
    luabind::object globalTable = luabind::globals(script);
    globalTable[moduleName]["InputSystem"] = &ScriptInputSystem::GetInstance();
}

#endif
