/*
 * OgreMax Sample Viewer and Scene Loader - Ogre3D-based viewer and code for loading and displaying .scene files
 * Copyright 2013 AND Entertainment
 *
 * This code is available under the OgreMax Free License:
 *   -You may use this code for any purpose, commercial or non-commercial.
 *   -If distributing derived works (that use this source code) in binary or source code form, 
 *    you must give the following credit in your work's end-user documentation: 
 *        "Portions of this work provided by OgreMax (www.ogremax.com)"
 *
 * AND Entertainment assumes no responsibility for any harm caused by using this code.
 * 
 * The OgreMax Sample Viewer and Scene Loader were released at www.ogremax.com 
 */


//Includes---------------------------------------------------------------------
#include "FlyingCamera.hpp"

using namespace OgreMax;
using namespace OgreMax::Types;


//Implementation---------------------------------------------------------------
FlyingCamera::FlyingCamera()
{
    this->upAxis = UP_AXIS_Y;
    this->cameraNode = 0;
    this->camera = 0;
    this->moveSpeed = 100;
}

FlyingCamera::~FlyingCamera()
{
}

void FlyingCamera::Initialize()
{
    if (this->cameraNode != 0)
    {
        this->position = this->cameraNode->_getDerivedPosition();
        this->rotation = this->cameraNode->_getDerivedOrientation();
    }
    else if (this->camera != 0)
    {
        this->position = this->camera->getRealPosition();
        this->rotation = this->camera->getRealOrientation();
    }
    else
    {
        this->position = Ogre::Vector3::ZERO;
        this->rotation = Ogre::Quaternion::IDENTITY;
    }
}

void FlyingCamera::SetUpAxis(UpAxis upAxis, bool setDefaultRotation)
{
    this->upAxis = upAxis;

    if (setDefaultRotation)
    {
        if (this->upAxis == UP_AXIS_Y)
            this->rotation = Ogre::Quaternion(1, 0, 0, 0);
        else
            this->rotation = Ogre::Quaternion((Ogre::Real)0.707107, (Ogre::Real)0.707107, (Ogre::Real)0, (Ogre::Real)0);
    }
}

const Ogre::Vector3& FlyingCamera::GetPosition() const
{
    return this->position;
}

void FlyingCamera::SetPosition(const Ogre::Vector3& position)
{
    this->position = position;
}

const Ogre::Quaternion& FlyingCamera::GetRotation()
{
    return this->rotation;
}

void FlyingCamera::SetRotation(const Ogre::Quaternion& rotation)
{
    this->rotation = rotation;
}

void FlyingCamera::SetCamera(Ogre::Camera* camera, bool useParentNodeIfAvailable, bool copySettings)
{
    Ogre::Node* node = camera->getParentNode();
    if (useParentNodeIfAvailable && node != 0)
        SetCameraNode(node, copySettings);
    else
    {
        this->camera = camera;
        this->cameraNode = 0;

        if (copySettings)
            Initialize();
    }
}

void FlyingCamera::SetCameraNode(Ogre::Node* cameraNode, bool copySettings)
{
    this->cameraNode = cameraNode;
    this->camera = 0;

    if (copySettings)
        Initialize();
}

void FlyingCamera::Move(const Ogre::Vector3& move, Ogre::Real elapsedTime)
{
    if (move != Ogre::Vector3::ZERO)
    {
        //Get camera directions
        Ogre::Vector3 rightDirection, forwardDirection, upDirection;
        GetDirections(rightDirection, forwardDirection, upDirection);

        Ogre::Vector3 offset = move * (this->moveSpeed * elapsedTime);

        //Move current position
        this->position +=
            rightDirection * offset.x +
            forwardDirection * offset.y +
            upDirection * offset.z;
    }
}

void FlyingCamera::LookAt(const Ogre::Vector3& targetPoint)
{
	const Ogre::Vector3 offset = targetPoint - this->position;
    if (offset != Ogre::Vector3::ZERO) 
	{
		Ogre::Vector3 negativeDirection = -offset;
		negativeDirection.normalise();

		const Ogre::Vector3& yawAxis = (this->upAxis == UP_AXIS_Y) ? Ogre::Vector3::UNIT_Y : Ogre::Vector3::UNIT_Z;
		Ogre::Vector3 xAxis = yawAxis.crossProduct(negativeDirection).normalisedCopy();    
		Ogre::Vector3 yAxis = negativeDirection.crossProduct(xAxis).normalisedCopy();
	    
		this->rotation.FromAxes(xAxis, yAxis, negativeDirection);
	}
}

void FlyingCamera::Orbit(const Ogre::Vector2& rotate, const Ogre::Vector3& targetPoint)
{
	if (rotate != Ogre::Vector2::ZERO)
    {
		LookAt(targetPoint);
		Rotate(rotate);

		float distance = (targetPoint - this->position).length();
		Ogre::Vector3 direction = this->rotation * Ogre::Vector3::UNIT_Z;
		this->position = targetPoint + direction * distance;
    }
}

void FlyingCamera::Rotate(const Ogre::Vector2& rotate)
{
    if (rotate != Ogre::Vector2::ZERO)
    {
        Ogre::Real tilt = rotate.y;
        Ogre::Real turn = -rotate.x;

        //Create tilt/turn quaternions
        Ogre::Quaternion tiltRotationQuat(Ogre::Radian(tilt), Ogre::Vector3::UNIT_X);
        Ogre::Quaternion turnRotationQuat;
        if (this->upAxis == UP_AXIS_Y)
            turnRotationQuat = Ogre::Quaternion(Ogre::Radian(turn), Ogre::Vector3::UNIT_Y);
        else
            turnRotationQuat = Ogre::Quaternion(Ogre::Radian(turn), Ogre::Vector3::UNIT_Z);

        //Calculate new orientation, tilting first (in local space), then turning (in world space)
        this->rotation = turnRotationQuat * this->rotation * tiltRotationQuat;
    }
}

void FlyingCamera::UpdateCamera()
{
    if (this->cameraNode != 0)
    {
        this->cameraNode->setPosition(this->position);
        this->cameraNode->setOrientation(this->rotation);
    }
    else if (this->camera != 0)
    {
        this->camera->setPosition(this->position);
        this->camera->setOrientation(this->rotation);
    }
}

void FlyingCamera::GetDirections(Ogre::Vector3& rightDirection, Ogre::Vector3& forwardDirection, Ogre::Vector3& upDirection)
{
    //Convert rotation to matrix
    Ogre::Matrix3 rotationMatrix;
    this->rotation.ToRotationMatrix(rotationMatrix);

    //Get appropriate column vectors
    rightDirection = rotationMatrix.GetColumn(0);
    forwardDirection = -rotationMatrix.GetColumn(2);

    //The up direction is special since it is in world space
    if (this->upAxis == UP_AXIS_Y)
        upDirection = Ogre::Vector3::UNIT_Y;
    else
        upDirection = Ogre::Vector3::UNIT_Z;
}

