/*
 * OgreMax Sample Viewer and Scene Loader - Ogre3D-based viewer and code for loading and displaying .scene files
 * Copyright 2013 AND Entertainment
 *
 * This code is available under the OgreMax Free License:
 *   -You may use this code for any purpose, commercial or non-commercial.
 *   -If distributing derived works (that use this source code) in binary or source code form, 
 *    you must give the following credit in your work's end-user documentation: 
 *        "Portions of this work provided by OgreMax (www.ogremax.com)"
 *
 * AND Entertainment assumes no responsibility for any harm caused by using this code.
 * 
 * The OgreMax Sample Viewer and Scene Loader were released at www.ogremax.com 
 */


//Includes---------------------------------------------------------------------
#include "Application.hpp"
#include "InputSystem.hpp"
#include "SceneViewerState.hpp"
#include "OgreMaxUtilities.hpp"
#include "ApplicationWindow.hpp"
#include "IState.hpp"
#include <OgreRoot.h>
#include <OgreWindowEventUtilities.h>
#include <OgreRenderSystem.h>
#include <OgreTextureManager.h>
#include <OgreConfigFile.h>

using namespace Ogre;
using namespace OgreMax;


//Local functions--------------------------------------------------------------
static bool FileExists(const String& fileName)
{
#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
    DWORD attributes = GetFileAttributes(fileName.c_str());
    return attributes != INVALID_FILE_ATTRIBUTES && (attributes & (FILE_ATTRIBUTE_NORMAL | FILE_ATTRIBUTE_ARCHIVE));
#else
    struct stat statbuf;
    return stat(fileName.c_str(), &statbuf) == 0;
#endif
}


//Static initialization--------------------------------------------------------
Application Application::instance;


//Implementation---------------------------------------------------------------
Application::Application()
{
    this->renderSystem = 0;
    this->applicationTitle = "OgreMax Sample Viewer";
    this->applicationName = "OgreMaxSampleViewer";
    this->hadFocus = false;
    this->suspendAllowed = false;
    this->toggleFullScreen = false;
    this->currentState = 0;
    this->elapsedTime = 0;
    this->resourcePath = OgreMaxUtilities::GetApplicationResourcesPath();
}

Application::~Application()
{    
}

void Application::SetSettings(const ApplicationSettings& settings)
{
    this->settings = settings;
}

const ApplicationSettings& Application::GetSettings() const
{
    return this->settings;
}

bool Application::Run(int argc, char** argv)
{
    CommandLineArgs commandLineArgs(argc, argv);
    ApplicationSettings settings(commandLineArgs);
    SetSettings(settings);
    return Run();
}

bool Application::Run()
{
    bool result = false;

    try
    {
        //Process the scene file name
        String fileName;
        StringUtil::splitFilename(this->settings.fileName, fileName, this->sceneDirectory);

        //Set up scene viewer state
        SceneViewerState& sceneViewerState = SceneViewerState::GetInstance();
        sceneViewerState.SetFileToLoad(fileName);
        sceneViewerState.SetLockViewCameraToSceneCamera(this->settings.lockViewCameraToSceneCamera);

        //Initialize
        Initialize();

        //Activate the start state
        IState& startState = sceneViewerState;
        startState.Activating();

        //The main loop
        this->currentState = &startState;
        do
        {
            this->hadFocus = this->applicationWindow->HasFocus();
            if (this->toggleFullScreen)
            {
                this->applicationWindow->ToggleFullScreen();
                this->toggleFullScreen = false;
            }

            //Suspend the application for a bit if the scene viewer state is suspended
            if (this->currentState == &SceneViewerState::GetInstance() &&
                SceneViewerState::GetInstance().IsSuspended())
            {
                #if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
                    WaitMessage();
                #endif
            }
            WindowEventUtilities::messagePump();

            //Manually fire frame started event
            Root::getSingleton()._fireFrameStarted();

            //Update only if the state wasn't deactivated by the user closing the window
            if (this->currentState != 0)
            {
                //Update the current state
                IState::UpdateState(this->currentState);

                //If not ending the program, draw the current frame and fire frame ended event
                if (this->currentState != 0)
                {
                    Root::getSingleton()._updateAllRenderTargets();
                    Root::getSingleton()._fireFrameEnded();
                }
            }
        }while (this->currentState != 0);

        result = true;
    }
    catch (Ogre::Exception e)
    {
        String title = this->applicationTitle;
        title += " Error";
        ApplicationWindow::ShowErrorMessage(e.getFullDescription().c_str(), title.c_str());
    }

    //Final cleanup
    Shutdown();

    return result;
}

void Application::Initialize()
{
    //Graphics system--------------------------------------
    String applicationDirectory = OgreMaxUtilities::GetApplicationDirectory();

    //Plugins configuration file name
    String pluginsConfigurationFilePath;
    if (!this->settings.pluginsConfigurationFileName.empty())
        pluginsConfigurationFilePath = OgreMaxUtilities::JoinPath(applicationDirectory, this->settings.pluginsConfigurationFileName);
    else
	    pluginsConfigurationFilePath = this->resourcePath + this->applicationName + "Plugins.cfg";
    if (!FileExists(pluginsConfigurationFilePath))
        pluginsConfigurationFilePath.clear();

    //Resources configuration file name
    String configurationFileName = this->resourcePath + this->applicationName + ".cfg";

    //Log file name
    this->logFileName = applicationDirectory + this->applicationName + ".log";

    //Create Ogre root
    new Root(pluginsConfigurationFilePath, configurationFileName, this->logFileName);

    //Instantiate the Ogre plugins
    if (pluginsConfigurationFilePath.empty())
        this->ogreStaticPlugins.Create();

    //Configure the render system
    this->renderSystem = this->settings.ConfigureRenderSystem();

    //Initialize Ogre
    Root::getSingleton().initialise(false);

    //Initialize the application/render window
    this->applicationWindow = new ApplicationWindow;
    ApplicationWindowSettings windowSettings(this->settings.screenWidth, this->settings.screenHeight, this->settings.fullScreen);
    windowSettings.enableNvidiaPerformanceHUD = this->settings.enableNvidiaPerformanceHUD;
    this->applicationWindow->Create(this->applicationTitle, this->applicationName, windowSettings);

    //Get rid of the garbage in the frame buffer so the user doesn't have to look at it if
    //no initial scene file is specified
    this->renderSystem->clearFrameBuffer(FBT_COLOUR | FBT_DEPTH | FBT_STENCIL);

    //Other graphics settings
    GpuNamedConstants::setGenerateAllConstantDefinitionArrayEntries(true);

    //Add the frame listener
    Root::getSingleton().addFrameListener(this);

    //Add the window event listener
    this->applicationWindow->SetWindowEventListener(this);

    //Input system-----------------------------------------
    this->hadFocus = true;
#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
    this->suspendAllowed = !this->settings.fullScreen;
#else
    this->suspendAllowed = false;
#endif
    InputSystem::GetInstance().Initialize((size_t)this->applicationWindow->GetWindowHandle(), !this->suspendAllowed);    
}

void Application::Shutdown()
{
    //Input system-----------------------------------------
    InputSystem::GetInstance().Shutdown();

    //Graphics system--------------------------------------
    delete Root::getSingletonPtr();
    this->renderSystem = 0;
    this->ogreStaticPlugins.Destroy();

    //Miscellaneous----------------------------------------
    //Destroy application window
    if (this->applicationWindow != 0)
    {
        this->applicationWindow->SetWindowEventListener(0);

        delete this->applicationWindow;
        this->applicationWindow = 0;
    }

    //Delete log file
    if (!this->settings.keepLogFile)
        OgreMaxUtilities::RemoveFile(this->logFileName);
}

bool Application::frameEnded(const FrameEvent& evt)
{
    this->elapsedTime = evt.timeSinceLastFrame;
    return true;
}

void Application::WindowResized()
{
	int width, height;
	int left, top;
	applicationWindow->GetWindowRectangle(left, top, width, height);

    if (!InputSystem::GetInstance().Mouse.IsNull())
    {
        const OIS::MouseState& ms = InputSystem::GetInstance().Mouse.GetMouse()->getMouseState();
	    ms.width = width;
	    ms.height = height;
    }
}

void Application::WindowClosing()
{
    if (this->currentState != 0)
    {
        this->currentState->Deactivating();
        this->currentState = 0;
    }

    this->applicationWindow = 0;
}

Application& Application::GetInstance()
{
    return instance;
}
