/*
 * OgreMax Sample Viewer and Scene Loader - Ogre3D-based viewer and code for loading and displaying .scene files
 * Copyright 2013 AND Entertainment
 *
 * This code is available under the OgreMax Free License:
 *   -You may use this code for any purpose, commercial or non-commercial.
 *   -If distributing derived works (that use this source code) in binary or source code form, 
 *    you must give the following credit in your work's end-user documentation: 
 *        "Portions of this work provided by OgreMax (www.ogremax.com)"
 *
 * AND Entertainment assumes no responsibility for any harm caused by using this code.
 * 
 * The OgreMax Sample Viewer and Scene Loader were released at www.ogremax.com 
 */


//Includes---------------------------------------------------------------------
#include <wx/frame.h>
#include <wx/display.h>
#include <wx/msgdlg.h>
#include <wx/icon.h>
#include "ApplicationWindow.hpp"
#include <OgreCommon.h>
#include <OgreStringConverter.h>
#include <OgreRoot.h>
#include <OgreRenderWindow.h>

using namespace Ogre;
using namespace OgreMax;


//Implementation---------------------------------------------------------------
ApplicationWindow::ApplicationWindow()
{
    this->window = 0;
    this->renderWindow = 0;
    this->listener = 0;
    this->hasFocus = false;    
}

ApplicationWindow::~ApplicationWindow()
{
    Destroy();
}

void ApplicationWindow::Create(const String& title, const String& internalName, const ApplicationWindowSettings& settings)
{
    //Make copy of settings
    this->settings = settings;
    wxDisplay display(0);
    wxRect desktopRect = display.GetGeometry();
    this->settings.LimitDimensions(desktopRect.width, desktopRect.height);

    //Create the window
    long style = wxSYSTEM_MENU | wxMINIMIZE_BOX | wxMAXIMIZE_BOX | wxCLOSE_BOX | wxCAPTION | wxCLIP_CHILDREN | wxNO_FULL_REPAINT_ON_RESIZE | wxRESIZE_BORDER;
    int width, height;    
    this->settings.GetActiveDimensions(width, height);
    this->window = new wxFrame(0, wxID_ANY, title, wxDefaultPosition, wxDefaultSize, style, internalName);
    this->window->SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    this->window->SetClientSize(width, height);
    this->window->CenterOnScreen();
    this->window->Show();
    this->hasFocus = true;
    
    //Set up event handlers    
    this->window->Connect(wxEVT_CLOSE_WINDOW, wxCloseEventHandler(ApplicationWindow::OnClose), 0, this);
    this->window->Connect(wxEVT_SIZE, wxSizeEventHandler(ApplicationWindow::OnSize), 0, this);
    this->window->Connect(wxEVT_ACTIVATE, wxActivateEventHandler(ApplicationWindow::OnActivate), 0, this);
        
    //Set window icon
    this->window->SetIcon(wxIcon(wxT("IDI_OGREMAX"), wxBITMAP_TYPE_ICO_RESOURCE));

    //Create Ogre render window
    this->renderWindowWidth = width;
    this->renderWindowHeight = height;
    Ogre::NameValuePairList renderWindowOptions;
    renderWindowOptions["externalWindowHandle"] = Ogre::StringConverter::toString((size_t)this->window->GetHandle());
    if (settings.enableNvidiaPerformanceHUD)
        renderWindowOptions["useNVPerfHUD"] = "true";
    this->renderWindow = Ogre::Root::getSingleton().createRenderWindow
        (
        internalName,
        this->renderWindowWidth,
        this->renderWindowHeight,
        this->settings.fullScreen,
        &renderWindowOptions
        );
}

void ApplicationWindow::Destroy()
{
    //No need to delete the render window since Ogre will destroy that automatically

    this->listener = 0;

    //Delete window
    if (this->window != 0)
    {
        this->window->Disconnect(wxEVT_CLOSE_WINDOW, wxCloseEventHandler(ApplicationWindow::OnClose), 0, this);
        this->window->Disconnect(wxEVT_SIZE, wxSizeEventHandler(ApplicationWindow::OnSize), 0, this);
        this->window->Disconnect(wxEVT_ACTIVATE, wxActivateEventHandler(ApplicationWindow::OnActivate), 0, this);

        this->window->Destroy();
        this->window = 0;

        this->renderWindow = 0;
    }
}

void ApplicationWindow::ShowMessage(const char* message, const char* title)
{
    wxMessageDialog dialog(this->window, message, title, wxOK | wxCENTER);
    dialog.ShowModal();
}

void ApplicationWindow::ShowErrorMessage(const char* message, const char* title)
{
    wxMessageDialog dialog(0, message, title, wxOK | wxCENTER | wxICON_ERROR);
    dialog.ShowModal();    
}

RenderWindow* ApplicationWindow::GetRenderWindow()  
{
    return this->renderWindow;
}

bool ApplicationWindow::HasFocus() const 
{
    return this->hasFocus;
}

wxWindow* ApplicationWindow::GetWindow() const 
{
    return this->window;
}

void* ApplicationWindow::GetWindowHandle() const 
{
    return (void*)this->window->GetHandle();
}

void ApplicationWindow::GetWindowRectangle(int& left, int& top, int& width, int& height)
{
    if (this->settings.fullScreen)
    {
        left = 0;
        top = 0;
        width = this->settings.fullScreenWidth;
        height = this->settings.fullScreenHeight;
    }
    else
    {
        wxRect rect = this->window->GetRect();    
        left = rect.x;
        top = rect.y;
        width = rect.width;
        height = rect.height;
    }
}

void ApplicationWindow::GetClientRectangle(int& left, int& top, int& width, int& height)
{
    if (this->settings.fullScreen)
    {
        left = 0;
        top = 0;
        width = this->settings.fullScreenWidth;
        height = this->settings.fullScreenHeight;
    }
    else
    {
        wxRect rect = this->window->GetClientRect();
        left = rect.x;
        top = rect.y;
        width = rect.width;
        height = rect.height;
    }
}

float ApplicationWindow::GetAspectRatio() const
{
    return (float)this->renderWindowWidth/this->renderWindowHeight;
}

void ApplicationWindow::ToggleFullScreen()
{
    this->settings.fullScreen = !this->settings.fullScreen;
    this->settings.GetActiveDimensions(this->renderWindowWidth, this->renderWindowHeight);
    this->renderWindow->setFullscreen(this->settings.fullScreen, this->renderWindowWidth, this->renderWindowHeight);
    if (this->listener != 0)
        this->listener->WindowResized();
}

void ApplicationWindow::SetWindowEventListener(ApplicationWindowEventListener* listener)
{
    this->listener = listener;
}

void ApplicationWindow::CenterCursor()
{
    //Get the render window rectangle
    unsigned int width, height, depth;
	int left, top;
	this->renderWindow->getMetrics(width, height, depth, left, top);

#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
    //Get the monitor for the window
    HMONITOR hMonitor = MonitorFromWindow((HWND)this->window->GetHandle(), MONITOR_DEFAULTTONEAREST);

    //Adjust the render window position based on the current monitor rectangle
    MONITORINFO monitorInfo;
    monitorInfo.cbSize = sizeof(monitorInfo);
    if (GetMonitorInfo(hMonitor, &monitorInfo))
    {
	    left += monitorInfo.rcMonitor.left;
        top += monitorInfo.rcMonitor.top;
    }
#endif

    //Set cursor position
    SetCursorPos(left + width/2, top + height/2);
}

void ApplicationWindow::OnClose(wxCloseEvent& event)
{
    if (this->listener != 0)
        this->listener->WindowClosing();
}

void ApplicationWindow::OnSize(wxSizeEvent& event)
{
    if (this->renderWindow != 0)
        this->renderWindow->windowMovedOrResized();
    
    wxRect rect = this->window->GetClientRect();
    this->renderWindowWidth = rect.width;
    this->renderWindowHeight = rect.height;

    if (this->listener != 0)
	    this->listener->WindowResized();
}

void ApplicationWindow::OnActivate(wxActivateEvent& event)
{
    this->hasFocus = event.GetActive();
}