/*
 * OgreMax Sample Viewer and Scene Loader - Ogre3D-based viewer and code for loading and displaying .scene files
 * Copyright 2013 AND Entertainment
 *
 * This code is available under the OgreMax Free License:
 *   -You may use this code for any purpose, commercial or non-commercial.
 *   -If distributing derived works (that use this source code) in binary or source code form, 
 *    you must give the following credit in your work's end-user documentation: 
 *        "Portions of this work provided by OgreMax (www.ogremax.com)"
 *
 * AND Entertainment assumes no responsibility for any harm caused by using this code.
 * 
 * The OgreMax Sample Viewer and Scene Loader were released at www.ogremax.com 
 */


#ifndef OgreMax_ApplicationWindow_INCLUDED
#define OgreMax_ApplicationWindow_INCLUDED


//Includes---------------------------------------------------------------------
#include <wx/event.h>
#include <OgreString.h>


//Classes----------------------------------------------------------------------
class wxFrame;
class wxCloseEvent;
class wxSizeEvent;
class wxActivateEvent;

namespace Ogre
{
    class RenderSystem;
    class RenderWindow;    
}

namespace OgreMax
{
    class IState;
    class ApplicationWindow;
    class CommandLineArgs;

    /** Event listener for the ApplicationWindow class */
    class ApplicationWindowEventListener
	{
	public:
		virtual ~ApplicationWindowEventListener() {};

        /** Called when the window is resized */
		virtual void WindowResized() {}

        /** Called just before the window is closed */
		virtual void WindowClosing()  {}
	};

    struct ApplicationWindowSettings
    {
        int width, height;
        int fullScreenWidth, fullScreenHeight;
        bool fullScreen;
        bool enableNvidiaPerformanceHUD;

        ApplicationWindowSettings()
        {
            this->width = 800;
            this->height = 600;
            this->fullScreenWidth = 800;
            this->fullScreenHeight = 600;
            this->fullScreen = false;
            this->enableNvidiaPerformanceHUD = false;
        }

        ApplicationWindowSettings(int width, int height, bool fullScreen)
        {
            this->width = width;
            this->height = height;
            this->fullScreenWidth = width;
            this->fullScreenHeight = height;
            this->fullScreen = fullScreen;
            this->enableNvidiaPerformanceHUD = false;
        }

        void LimitDimensions(int maxWidth, int maxHeight)
        {
            if (this->width > maxWidth) this->width = maxWidth;
            if (this->height> maxHeight) this->height = maxHeight;
            if (this->fullScreenWidth > maxWidth) this->fullScreenWidth = maxWidth;
            if (this->fullScreenHeight > maxHeight) this->fullScreenHeight = maxHeight;
        }

        void GetActiveDimensions(int& width, int& height) const
        {
            if (this->fullScreen)
            {
                width = this->fullScreenWidth;
                height = this->fullScreenHeight;
            }
            else
            {
                width = this->width;
                height = this->height;
            }
        }
    };

    class ApplicationWindow : public wxEvtHandler
    {
    public:
        ApplicationWindow();
        virtual ~ApplicationWindow();

        void Create(const Ogre::String& title, const Ogre::String& internalName, const ApplicationWindowSettings& settings);
        void Destroy();

        /** Displays a message box that has the window as its parent. */
        void ShowMessage(const char* message, const char* title);

        /** Displays an error message that has the desktop as its parent. */
        static void ShowErrorMessage(const char* message, const char* title);

        /** Gets the window's Ogre RenderWindow. */
        Ogre::RenderWindow* GetRenderWindow();

        /** Determines if the window has input focus. */
        bool HasFocus() const;

        /** Gets the operating system window handle for the window. */
        wxWindow* GetWindow() const;

        /** Gets the operating system window handle for the window. */
        void* GetWindowHandle() const;

        /** Gets the bounding rectangle for the window. */
        void GetWindowRectangle(int& left, int& top, int& width, int& height);

        /** Gets the rectangle for the client area. */
        void GetClientRectangle(int& left, int& top, int& width, int& height);

        /** Gets the client render window aspect ratio. */
        float GetAspectRatio() const;

        /** Switches back and forth between full screen and windowed mode. */
        void ToggleFullScreen();

        /** Adds a window event listener. */
        void SetWindowEventListener(ApplicationWindowEventListener* listener);

        /** Centers the cursor in the application window */
        void CenterCursor();

    private:
        void OnClose(wxCloseEvent& event);
        void OnSize(wxSizeEvent& event);
        void OnActivate(wxActivateEvent& event);

    protected:
        /** The operating system window handle */
        wxFrame* window;

        /** The Ogre render window */
        Ogre::RenderWindow* renderWindow;

        ApplicationWindowEventListener* listener;
        
        /** Indicates whether the window has input focus */
        bool hasFocus;

        ApplicationWindowSettings settings;

        //Window dimensions
        int renderWindowWidth;
        int renderWindowHeight;
    };
}


#endif
