/*
 * OgreMax Sample Viewer and Scene Loader - Ogre3D-based viewer and code for loading and displaying .scene files
 * Copyright 2013 AND Entertainment
 *
 * This code is available under the OgreMax Free License:
 *   -You may use this code for any purpose, commercial or non-commercial.
 *   -If distributing derived works (that use this source code) in binary or source code form, 
 *    you must give the following credit in your work's end-user documentation: 
 *        "Portions of this work provided by OgreMax (www.ogremax.com)"
 *
 * AND Entertainment assumes no responsibility for any harm caused by using this code.
 * 
 * The OgreMax Sample Viewer and Scene Loader were released at www.ogremax.com 
 */


//Includes---------------------------------------------------------------------
#include "SceneViewerState.hpp"
#ifdef _Char
    #undef _Char
#endif
#include "InputSystem.hpp"
#include "OgreMaxUtilities.hpp"
#include "ViewerConstants.hpp"
#include "ApplicationUtilities.hpp"
#include "Application.hpp"
#include <OgreRenderWindow.h>
#include <OgreRoot.h>

using namespace Ogre;
using namespace OgreMax;
using namespace OgreMax::Types;


//Static initialization--------------------------------------------------------
SceneViewerState SceneViewerState::instance;


//Local functions--------------------------------------------------------------
/** Shows or hides the cursor. Used only under Windows */
static void ShowTheCursor(bool show)
{
#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
    ShowCursor(show);
#endif
}


//Implementation---------------------------------------------------------------
SceneViewerState::SceneViewerState()
{
    this->scene = 0;
    this->sceneManager = 0;
    this->upAxis = UP_AXIS_Y;

    this->currentCameraIndex = -1;
    this->viewerCamera = 0;
    this->viewerReflectionCamera = 0;
    this->lockViewCameraToSceneCamera = false;
    this->suspended = false;
}

SceneViewerState::~SceneViewerState()
{
}

SceneViewerState& SceneViewerState::GetInstance()
{
    return instance;
}

void SceneViewerState::SetFileToLoad(const String& fileToLoad)
{
    this->fileToLoad = fileToLoad;
}

void SceneViewerState::SetLockViewCameraToSceneCamera(bool lock)
{
    this->lockViewCameraToSceneCamera = lock;
}

void SceneViewerState::Activating()
{
    Activate(true);
}

void SceneViewerState::Deactivating()
{
    Deactivate(true);
}

IState* SceneViewerState::Update()
{
    Application& app = Application::GetInstance();

    IState* nextState = this;

    //Determine if the program should be suspended
    if (app.GainedFocus())
        Suspend(false);
    else if (app.LostFocus())
        Suspend(true);
    
    if (app.HasFocus())
    {
        //Update the input system
        InputSystem::GetInstance().Update();

        if (this->sceneManager == 0)
        {
            //No scene is loaded

            //Get user input events
            UserEvents userEvents;
            this->userActionMap.GetEvents(userEvents);

            //Exit program?
            if (userEvents.ConsumeEvent(UserEventID::ESCAPE) || userEvents.ConsumeEvent(UserEventID::EXIT))
                nextState = 0;
        }
        else if (this->suspended)
        {
            //The program is suspended. The user can exit either the program or the suspended state

            //Get user input events
            SuspendedUserEvents userEvents;
            this->suspendedUserActionMap.GetEvents(userEvents);

            if (userEvents.ConsumeEvent(SuspendedUserEventID::ESCAPE))
            {
                //Exit program
                nextState = 0;
            }
            else if (userEvents.ConsumeEvent(SuspendedUserEventID::UNESCAPE))
            {
                //Exit suspended state
                Suspend(false);
            }
        }
        else
        {
            //The program is in its normal state

            //Get user input events
            UserEvents userEvents;
            this->userActionMap.GetEvents(userEvents);

            //Center the cursor in the application window
            app.GetApplicationWindow()->CenterCursor();

            //F12 is the non-overrideable full screen toggle key
            if (InputSystem::GetInstance().Keyboard.PressedKey(OIS::KC_F12))
                app.RequestToggleFullScreen();

            //Change current camera
            if (userEvents.ConsumeEvent(UserEventID::CHANGE_CAMERA) && !this->cameras.empty())
            {
                if (userEvents.cameraIndex != -1)
                    SetCurrentCamera(userEvents.cameraIndex);
                else if (userEvents.cameraIndexDirection != 0)
                    SetPreviousOrNextCamera(userEvents.cameraIndexDirection);
            }

            //Move/rotate camera
            //Move camera
            if (userEvents.ConsumeEvent(UserEventID::MOVE))
            {
                this->flyingCamera.Move(userEvents.move, app.GetElapsedTime());
                this->lockViewCameraToSceneCamera = false;
            }

            //Rotate camera
            if (userEvents.ConsumeEvent(UserEventID::LOOK))
            {
                this->flyingCamera.Rotate(userEvents.look);
                this->lockViewCameraToSceneCamera = false;
            }

            //Toggle view camera locking
            if (userEvents.ConsumeEvent(UserEventID::TOGGLE_VIEW_CAMERA_LOCK))
                this->lockViewCameraToSceneCamera = !this->lockViewCameraToSceneCamera;

            //Capture a screenshot
            if (userEvents.ConsumeEvent(UserEventID::SCREENSHOT))
            {
                const String& baseName = app.GetSettings().screenshotBaseName;
                const String& ext = app.GetSettings().screenshotExtension;
                app.GetRenderWindow()->writeContentsToTimestampedFile(baseName, ext);
            }

            //Suspend or exit program?
            if (userEvents.ConsumeEvent(UserEventID::ESCAPE))
            {
                if (!Suspend(true))
                {
                    //Suspend not allowed. Exit program
                    nextState = 0;
                }
            }

            //Exit program?
            if (userEvents.ConsumeEvent(UserEventID::EXIT))
                nextState = 0;

            //Update view camera
            if (!this->cameras.empty() && this->lockViewCameraToSceneCamera)
                SetCurrentCamera(this->currentCameraIndex);
            else
                this->flyingCamera.UpdateCamera();

            //Update reflection camera
            ApplicationUtilities::CopyCameraSettings(this->viewerReflectionCamera, this->viewerCamera);

            //Update scene
            const Real elapsedTime = app.GetElapsedTime();
            if (this->scene != 0)
                this->scene->Update(elapsedTime);
        }
    }

    return nextState;
}

void SceneViewerState::Activate(bool starting)
{
    Application& app = Application::GetInstance();

    if (starting)
    {
        //Initialize user action map
        this->userActionMap.Initialize(ViewerConstants::USER_CONTROLS_FILE_NAME);

        //Hide the cursor if there is a scene, otherwise flag the scene as suspended
        if (!this->fileToLoad.empty())
            ShowTheCursor(false);
        else
            this->suspended = true;

        //Set the camera movement speed
        this->flyingCamera.SetMoveSpeed(app.GetSettings().cameraMoveSpeed);
    }

    //Exit early if the scene hasn't been loaded
    if (this->fileToLoad.empty())
        return;

    //Create the scene
    Vector3 cameraPosition = Vector3::ZERO;
    CreateScene();

    //Initialize the flying camera
    if (this->cameras.empty())
    {
        //There are no scene cameras. Set up the flying camera manually
        this->viewerCamera->setPosition(cameraPosition);
        this->flyingCamera.SetCamera(this->viewerCamera);
        this->flyingCamera.SetUpAxis(this->upAxis, true);
    }
    else
    {
        //A scene camera will be used. Set the up axis and nothing else
        this->flyingCamera.SetUpAxis(this->upAxis, false);
    }
}

void SceneViewerState::Deactivate(bool exiting)
{
    Application& app = Application::GetInstance();

    if (exiting)
    {
        //Save the user control mappings
        if (app.GetSettings().writeUserControlsFile)
            this->userActionMap.SaveFile(ViewerConstants::USER_CONTROLS_FILE_NAME);

        //Show the cursor
        ShowTheCursor(true);
    }

    //Exit early if the scene hasn't been loaded
    if (this->sceneManager == 0)
        return;

    //Clear the scene
    if (app.GetRenderWindow() != 0)
        app.GetRenderWindow()->removeAllViewports();
    delete this->scene;
    this->scene = 0;
    this->sceneManager->clearScene();
    Root::getSingleton().destroySceneManager(this->sceneManager);
    this->sceneManager = 0;

    //Reset some other members
    this->upAxis = UP_AXIS_Y;
    this->cameras.clear();
    this->currentCameraIndex = -1;
    this->viewerCamera = 0;
    this->viewerReflectionCamera = 0;

    //Destroy and recreate the necessary resource groups
    OgreMaxUtilities::ResetResourceGroups(Ogre::ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
}

void SceneViewerState::SetCurrentCamera(int cameraIndex)
{
    if (cameraIndex >= 0 && cameraIndex < (int)this->cameras.size())
    {
        this->currentCameraIndex = cameraIndex;

        ApplicationUtilities::CopyCameraSettings(this->viewerCamera, this->cameras[this->currentCameraIndex]);
        this->flyingCamera.SetCamera(this->viewerCamera);
    }
}

void SceneViewerState::SetPreviousOrNextCamera(int direction)
{
    //Make sure direction is valid
    direction = (direction < 0) ? -1 : 1;

    //Set new camera
    int cameraCount = (int)this->cameras.size();
    SetCurrentCamera((this->currentCameraIndex + direction + cameraCount) % cameraCount);
}

bool SceneViewerState::Suspend(bool suspend)
{
    if (this->sceneManager != 0 && Application::GetInstance().SuspendAllowed())
    {
        //Show/hide cursor if suspend state is changing
        if (suspend != this->suspended)
            ShowTheCursor(suspend);

        this->suspended = suspend;

        return true;
    }
    else
        return false;
}

void SceneViewerState::CreateScene()
{
    Application& app = Application::GetInstance();

    //Create scene object
    this->scene = new OgreMaxScene;

    //Set the scene's base resource location and initialize necessary resource groups
    this->scene->SetBaseResourceLocation(app.GetSceneDirectory());

    //Assume it's a scene file
    try
    {
        this->scene->Load
            (
            this->fileToLoad,
            app.GetRenderWindow(),
            OgreMaxScene::SET_DEFAULT_LIGHTING,
            0,
            0,
            this
            );
    }
    catch (Exception ex)
    {            
        wxString message = "There was an error loading the scene: ";
        message += ex.getDescription();
        
        app.GetApplicationWindow()->ShowMessage(message, "OgreMax Sample Viewer Error");        
    }

    this->upAxis = this->scene->GetUpAxis();
}

void SceneViewerState::HandleNewSceneManager(SceneManager* sceneManager)
{
    //Keep the scene manager
    this->sceneManager = sceneManager;

    //Create the camera that the user is able to move around
    this->viewerCamera = sceneManager->createCamera("__viewerCamera");
    this->viewerReflectionCamera = sceneManager->createCamera("__viewerReflectionCamera");

    //Set default clip distances
    const Real nearClip = (Real).1;
    const Real farClip = (Real)100000;
    this->viewerCamera->setNearClipDistance(nearClip);
    this->viewerCamera->setFarClipDistance(farClip);

    //Create viewport
    Viewport* viewport = Application::GetInstance().GetRenderWindow()->addViewport(this->viewerCamera);

    //Alter the camera aspect ratio to match the viewport
    this->viewerCamera->setAspectRatio(Real(viewport->getActualWidth()) / Real(viewport->getActualHeight()));
}

void SceneViewerState::CreatedSceneManager(const OgreMaxScene* scene, SceneManager* sceneManager)
{
    HandleNewSceneManager(sceneManager);
}

void SceneViewerState::CreatedCamera(const OgreMaxScene* scene, Camera* camera)
{
    //Hold onto the camera
    this->cameras.push_back(camera);

    //Select the camera if it hasn't already been selected
    if (this->currentCameraIndex == -1)
    {
        this->currentCameraIndex = 0;
        SetCurrentCamera(this->currentCameraIndex);
    }
}

Camera* SceneViewerState::GetRenderTextureCamera
    (
    const OgreMaxScene* scene,
    const OgreMax::Types::RenderTextureParameters& renderTextureParameters
    )
{
    if (renderTextureParameters.textureType == TEX_TYPE_2D && renderTextureParameters.cameraName.empty())
    {
        //2D render texture in use and the camera name is empty. Use the viewer reflection camera
        return this->viewerReflectionCamera;
    }
    else
    {
        //Cube map render texture in use or camera name is empty. Use the scene camera with the specified name
        return 0;
    }
}

