/*
 * OgreMax Sample Viewer and Scene Loader - Ogre3D-based viewer and code for loading and displaying .scene files
 * Copyright 2013 AND Entertainment
 *
 * This code is available under the OgreMax Free License:
 *   -You may use this code for any purpose, commercial or non-commercial.
 *   -If distributing derived works (that use this source code) in binary or source code form, 
 *    you must give the following credit in your work's end-user documentation: 
 *        "Portions of this work provided by OgreMax (www.ogremax.com)"
 *
 * AND Entertainment assumes no responsibility for any harm caused by using this code.
 * 
 * The OgreMax Sample Viewer and Scene Loader were released at www.ogremax.com 
 */


#ifndef OgreMax_SceneViewerState_INCLUDED
#define OgreMax_SceneViewerState_INCLUDED


//Includes---------------------------------------------------------------------
#include "UserActionMap.hpp"
#include "SuspendedUserActionMap.hpp"
#include "FlyingCamera.hpp"
#include "OgreMaxScene.hpp"
#include "IState.hpp"


//Classes----------------------------------------------------------------------
namespace OgreMax
{
    class SceneViewerState :
        public IState,
        public OgreMaxSceneCallback
    {
    public:
        SceneViewerState();
        ~SceneViewerState();

        static SceneViewerState& GetInstance();

        /**
         * Sets the scene file to be viewed. The scene wil be loaded when Activating() is called
         */
        void SetFileToLoad(const Ogre::String& fileToLoad);

        /** Sets the view camera locking state */
        void SetLockViewCameraToSceneCamera(bool lock);

        /** Determines if the state is currently suspended */
        bool IsSuspended() const {return this->suspended;}

    protected:
        void Activate(bool starting);
        void Deactivate(bool exiting);

        /** Sets the current camera by index */
        void SetCurrentCamera(int cameraIndex);

        /** Sets the current camera by moving */
        void SetPreviousOrNextCamera(int direction);

        /**
         * Suspends the viewer state
         * @param suspend [in] - Indicates whether to suspend (true) or unsuspend (false)
         * @return Returns true if the suspend was allowed, false otherwise. The suspend won't be
         * allowed if the application settings prevent suspends from taking place
         */
        bool Suspend(bool suspend);

        void CreateScene();
        
        void HandleNewSceneManager(Ogre::SceneManager* sceneManager);

        //IState methods
        void Activating();
        void Deactivating();
        IState* Update();

        //OgreMaxSceneCallback callbacks
        void CreatedSceneManager(const OgreMax::OgreMaxScene* scene, Ogre::SceneManager* sceneManager);
        void CreatedCamera(const OgreMaxScene* scene, Ogre::Camera* camera);
        Ogre::Camera* GetRenderTextureCamera
            (
            const OgreMax::OgreMaxScene* scene,
            const OgreMax::Types::RenderTextureParameters& renderTextureParameters
            );        

    public:
        /**
         * File to view
         * This must first be set with a call to SetFileToLoad()
         */
        Ogre::String fileToLoad;

        /** The OgreMax scene. This will be null if a mesh file is being viewed. */
        OgreMaxScene* scene;

        /** The scene up axis */
        Types::UpAxis upAxis;

        /** The scene manager created by the OgreMaxScene. */
        Ogre::SceneManager* sceneManager;

        typedef std::vector<Ogre::Camera*> Cameras;

        /** All the cameras */
        Cameras cameras;

        //The camera used to view the scene
        int currentCameraIndex;
        Ogre::Camera* viewerCamera;
        Ogre::Camera* viewerReflectionCamera;
        FlyingCamera flyingCamera;

        /**
         * Indicates whether the view camera's transform should be locked to
         * the currently selected camera's transform
         */
        bool lockViewCameraToSceneCamera;

        /**
         * Indicates whether the program has been suspended, which occurs when the user
         * presses escape once in windowed mode
         */
        bool suspended;

        /** Maps user inputs to actions. Operates in "normal" (unsuspended) mode */
        UserActionMap userActionMap;

        /** Maps user inputs to actions. Operates in suspended mode */
        SuspendedUserActionMap suspendedUserActionMap;

        //The one and only instance of this class
        static SceneViewerState instance;
    };

}


#endif
